﻿using System;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Media;
using System.Windows.Media.Imaging;
using System.Windows.Shapes;
using System.Windows.Input;
using System.ComponentModel;
using System.Globalization;

namespace FastGaussFilterForRealImages
{ public partial class MainPage : UserControl
  { WriteableBitmap img0, img1, kernelImage;
    Int32 avSize, havSize;  //kernel width  and half kernel width,  kw uneven and hkw even
    double[] sigmas = { 1.6, 2.7, 3.8, 4.9, 6.0, 7.0, 8.2, 9.3, 10.4, 11.4, 12.5, 13.6, 14.8, 16.0, 17.3, 18.6, 19.8, 20.9, 21.8, 22.6 };
    Byte[,] kernel;
    Int32 xSize, ySize; //image width and height
    Byte [,] R0, G0, B0; //input
    Int32[,] R1, G1, B1; //intermediary
    Int32[,] R2, G2, B2; //intermediary
    Int32[,] R3, G3, B3; //output
    Int64 t0, t1;       //system time before and after filtering
    String fileName;
    BackgroundWorker myThread = new BackgroundWorker(); //Thread object
    CultureInfo international = new CultureInfo("en-US");

    public MainPage()
    { InitializeComponent();
      myThread.DoWork += new DoWorkEventHandler( myThreadDoWork    );
      myThread.RunWorkerCompleted += 
             new RunWorkerCompletedEventHandler( myThreadCompleted );
    }

    private void myImageIn_ImageOpened( object sender,RoutedEventArgs e ) 
    { img0  = new WriteableBitmap( (BitmapSource)myImageIn.Source );
      fileName = ((BitmapImage)myImageIn.Source).UriSource.OriginalString;
      a_new_image_opened();
    }

    private void myButtonOF_Click( object sender, RoutedEventArgs e )
    { OpenFileDialog openFileDialog = new OpenFileDialog();
      openFileDialog.Filter = "Image Files (*.png,*.jpg)|*.png;*.jpg";
      bool? userClickedOK = openFileDialog.ShowDialog();
      if ( userClickedOK == true )
      { try
        { System.IO.Stream stream = openFileDialog.File.OpenRead();
          fileName = openFileDialog.File.ToString();
          BitmapImage bmp = new BitmapImage();
          bmp.SetSource( stream );
          stream.Close();
          img0 = new WriteableBitmap( bmp );
          a_new_image_opened();
        } catch {}
      }
    }

    private void a_new_image_opened()
    { xSize = img0.PixelWidth;
      ySize = img0.PixelHeight;
      img1 = new WriteableBitmap( xSize, ySize );
      myImageIn    .Source = img0;
      myImageOut   .Source = img1;
      //prepare_intermediary_arrays
      R0 = new Byte [ySize, xSize];
      G0 = new Byte [ySize, xSize];
      B0 = new Byte [ySize, xSize];
      R1 = new Int32[ySize, xSize];
      G1 = new Int32[ySize, xSize];
      B1 = new Int32[ySize, xSize];
      R2 = new Int32[ySize, xSize];
      G2 = new Int32[ySize, xSize];
      B2 = new Int32[ySize, xSize];
      R3 = new Int32[ySize, xSize];
      G3 = new Int32[ySize, xSize];
      B3 = new Int32[ySize, xSize];
      GC.Collect(); //clean up memory when new arrays have replaced old ones
      for ( int y=0; y < ySize; y++ )
        for ( int x=0; x < xSize; x++ )
        { Int32 c = img0.Pixels[y*xSize+x];
          R0[y,x] = (Byte)( c & 0x0000ff)       ;
          G0[y,x] = (Byte)((c & 0x00ff00) >>  8);
          B0[y,x] = (Byte)((c & 0xff0000) >> 16);
        }
      avSize_Slider.Maximum = Math.Min( 61, xSize/4 ); //default 61, small images less
      avSize_Slider   .Value = 11; //initial kernel width
      myTextBoxOF .Text = fileName + ": " + xSize.ToString() + 'x' + ySize.ToString();
      myThreadStart();
    }

    private void myThreadStart()
    { if ( myThread.IsBusy ) return;
      myTextBoxTime.Text = "Wait !";
      myBorder.Cursor = Cursors.Wait;
      avSize = (Int32)avSize_Slider.Value; if (avSize%2 == 0) avSize--; avSizeTextBox.Text = avSize.ToString();
      havSize = avSize / 2;
      //Gaussian bell to kernel - 
      double stdDev = sigmas[havSize-1];
      stdDevTextBox.Text = stdDev.ToString( "0.0", international );
      double steepness = -1.0 / (2.0 * stdDev * stdDev);
      int GaussSize = Convert.ToInt32( 2*Math.Sqrt( stdDev*stdDev*( Math.Log(100) ) ) );
      if ( GaussSize % 2 == 0 ) GaussSize -= 1;
      int hGaussSize = GaussSize/2;
      kernel = new Byte[GaussSize,GaussSize];
      for ( int y=-hGaussSize; y <= hGaussSize; y++ ) //bell to kernel 
        for ( int x=-hGaussSize; x <= hGaussSize; x++ )
        { double vote = 100 * Math.Exp( steepness*(x*x + y*y) );
          kernel[y+hGaussSize,x+hGaussSize] = Convert.ToByte( vote );
        }
      //visualize kernel
      kernelImage = new WriteableBitmap( GaussSize, GaussSize );
      myImageKernel.Source = kernelImage;
      myImageKernel.Width  = myImageKernel.Height = GaussSize;
      for ( int y=0; y < GaussSize; y++ ) 
        for ( int x=0; x < GaussSize; x++ )
        { Byte p = Convert.ToByte( kernel[y,x] * 2.55f );
          unchecked { kernelImage.Pixels[y*GaussSize+x] = 
                      (Int32)0xff000000 | p | (p << 8) | (p << 16); }
        }
      kernelImage.Invalidate();
      t0 = DateTime.Now.Ticks; 
      myThread.RunWorkerAsync(); //Calls myThreadDoWork( ... )
    }

    private void myThreadDoWork( Object sender, DoWorkEventArgs e )
    { fastAverageFilter( 1 );
    }

    private void myThreadCompleted( Object sender,	RunWorkerCompletedEventArgs e )
    { myBorder.Cursor = Cursors.Arrow;
      t1 = DateTime.Now.Ticks;
      myTextBoxTime.Text = "This filter took " + ((t1-t0)/10000).ToString() + " millisec.";
      img1.Invalidate();
    }

    private void Slider_MouseLeftButtonUp( object sender, MouseButtonEventArgs e ) 
    { myThreadStart();
    }

    /*private void GaussFilter()
    { int R, G, B, Rsum, Gsum, Bsum, x, xx, xxx, y, yy, yyy, startOfYRow;
      int votes, sumOfVotes;
      for ( y=0; y < ySize; y++ ) //==================
      { startOfYRow = y*xSize;
        for ( x=0; x < xSize; x++ ) //===============
        { Rsum=Gsum=Bsum=sumOfVotes=0;
          for ( yy=-havSize; yy <= havSize; yy++ ) //=============
          { yyy = y + yy;
            if ( yyy < 0 | yyy >= ySize ) continue; //beyond border
            for ( xx=-havSize; xx <= havSize; xx++ )//========
            { xxx = x + xx;
              if ( xxx < 0 | xxx >= xSize ) continue; //beyond border
              sumOfVotes += votes = kernel[yy+havSize,xx+havSize];
              Rsum += R0[yyy,xxx] * votes;
              Gsum += G0[yyy,xxx] * votes;
              Bsum += B0[yyy,xxx] * votes;
            } //====== end for (int xx... ================
          } //======== end for (int yy... ==================
          R=Convert.ToInt32( (float)Rsum/sumOfVotes );      
          G=Convert.ToInt32( (float)Gsum/sumOfVotes ) <<  8;
          B=Convert.ToInt32( (float)Bsum/sumOfVotes ) << 16;
          unchecked { img1.Pixels[startOfYRow+x] = (Int32)0xff000000 | R | G | B; }
        } //============ end for (int x... =====================
      } //============== end for (int y... =======================
    }*/

    private void fastAverageFilter( int repetitions )
    { Int32 x, x0, x1, x2, x3, y, y0, y1, y2, y3;
      for ( y=0; y < ySize; y++ )
        for ( x=0; x < xSize; x++ ) 
        { R1[y,x] = R0[y,x];
          G1[y,x] = G0[y,x];
          B1[y,x] = B0[y,x];
        }     
      for ( int i0=0, i1=1; i0 < repetitions; i0++, i1++ )
      { for ( y=0; y < ySize; y++ )         //clear R2,G2, B2, R3, G3, B3
          for ( x=0; x < xSize; x++ ) 
            R2[y,x]=G2[y,x]=B2[y,x]=R3[y,x]=G3[y,x]=B3[y,x] = 0;
        //Horizontal summation into the intermediary integer-matrices R2, G2, B2
        for ( y=i0*havSize; y < ySize-i0*havSize; y++ )
        { //add up the first column = column[hkw]
          for ( x=i0*havSize; x < i0*havSize+avSize; x++ ) 
          { R2[y,havSize] += R1[y,x];
            G2[y,havSize] += G1[y,x];
            B2[y,havSize] += B1[y,x];
          }
          //add up all further columns from x2=[hkw+1] to x2=[xSize-hkw-1]
          for ( x0=i0*havSize, x1=i1*havSize, x2=i1*havSize+1, x3=i1*avSize; x3 < xSize-i0*havSize; x0++, x1++, x2++, x3++ )
          { R2[y,x2] = R2[y,x1] - R1[y,x0] + R1[y,x3];
            G2[y,x2] = G2[y,x1] - G1[y,x0] + G1[y,x3];
            B2[y,x2] = B2[y,x1] - B1[y,x0] + B1[y,x3];
          }
        }
        //Vertical summation into the output integer-matrices R3, G3, B3
        for ( x=i1*havSize; x < xSize-i1*havSize; x++ ) //vertical sums
        { for ( y=i0*havSize; y < i0*havSize+avSize; y++ )        //first row
          { R3[havSize,x] += R2[y,x];
            G3[havSize,x] += G2[y,x];
            B3[havSize,x] += B2[y,x];
          }
          for ( y0=i0*havSize, y1=i1*havSize, y2=i1*havSize+1, y3=i1*avSize; y3 < ySize-i0*havSize; y0++, y1++, y2++, y3++ )
          { R3[y2,x] = R3[y1,x] - R2[y0,x] + R2[y3,x];
            G3[y2,x] = G3[y1,x] - G2[y0,x] + G2[y3,x];
            B3[y2,x] = B3[y1,x] - B2[y0,x] + B2[y3,x];
          }
        }
        if ( i1 >= repetitions ) break; 
        for ( y=i1*avSize; y < ySize-i1*havSize; y++ )
          for ( x=i1*avSize; x < xSize-i1*havSize; x++ ) 
          { R1[y,x] = R3[y,x];
            G1[y,x] = G3[y,x];
            B1[y,x] = B3[y,x];
          }
      } //end of repetitions
      float divisor = (float)Math.Pow( avSize*avSize, repetitions );
      int yy, R, G, B;     
      for ( y=0; y < ySize; y++ )   //divide, round, shift and or together
      { yy = y * xSize;
        for ( x=0; x < xSize; x++ ) 
        { R = Convert.ToInt32( (float)R3[y,x] / divisor );
          G = Convert.ToInt32( (float)G3[y,x] / divisor ) <<  8;
          B = Convert.ToInt32( (float)B3[y,x] / divisor ) << 16;
          unchecked { img1.Pixels[yy+x] = (Int32)0xff000000 | B | G | R; }
        }
      }
    }
  }
}